import inspect
import marshal
import boto3
import json
from .picklelib import loads, dumps
import sys
import requests
import os
import botocore


class Chromeless():
    REQUIRED_SERVER_VERSION = 2

    def __init__(self, gateway_url=None, gateway_apikey=None, chrome_options=None,
                 function_name='chromeless-server-prod', boto3_session=None):
        self.boto3_session = boto3_session if boto3_session is not None else boto3
        self.gateway_url = gateway_url
        self.gateway_apikey = gateway_apikey
        self.options = chrome_options
        if function_name == 'chromeless-server-prod' and 'CHROMELESS_SERVER_FUNCTION_NAME' in os.environ:
            function_name = os.environ['CHROMELESS_SERVER_FUNCTION_NAME']
        self.function_name = function_name
        self.codes = {}

    def attach(self, method):
        try:
            self.codes[method.__name__] = inspect.getsource(
                method), marshal.dumps(method.__code__)
        except OSError as e:
            if "could not get source code" in str(e):
                raise RuntimeError(
                    "Chromeless does not support interactive mode. Please run from files.")
            else:
                raise e

    def __getattr__(self, name):
        if name in self.codes:
            self.invoked_func_name = name
            return self.__invoke
        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute {name}")

    def __invoke(self, *arg, **kw):
        dumped = dumps({
            "invoked_func_name": self.invoked_func_name,
            "codes": self.codes,
            "arg": arg,
            "kw": kw,
            "options": self.options,
            "REQUIRED_SERVER_VERSION": self.REQUIRED_SERVER_VERSION,
        })
        if self.function_name == "local":
            method = self.__invoke_local
        elif self.gateway_url is not None:
            method = self.__invoke_api
        else:
            method = self.__invoke_lambda
        response, metadata = loads(method(dumped))
        if metadata['status'] == "error":
            raise Exception(response)
        else:
            return response

    def __invoke_api(self, dumped):
        headers = {'x-api-key': self.gateway_apikey}
        return requests.post(self.gateway_url, headers=headers,
                             json={'dumped': dumped}).json()['result']

    def __invoke_local(self, dumped):
        response = requests.post(
            "http://"+os.environ['LOCAL_CHROMELESS_HOSTNAME']+":8080/2015-03-31/functions/function/invocations", json={'dumped': dumped})
        return response.text

    def __invoke_lambda(self, dumped):
        client = self.boto3_session.client('lambda')
        try:
            response = client.invoke(
                FunctionName=self.function_name,
                InvocationType='RequestResponse',
                LogType='Tail',
                Payload=json.dumps({'dumped': dumped})
            )
        except botocore.exceptions.ClientError as e:
            raise Exception(
                "Invalid session or AWS credentials: {}".format(str(e)))
        except Exception as e:
            raise
        return response['Payload'].read().decode()
